library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

use work.ipbus.all;

package pack_simu_transactor_lpsc is
type type_ipbus_buffer is array (natural range <>) of std_logic_vector(31 downto 0);

procedure ipbus_init(
				signal clk: in std_logic; 
				signal cntrl_to_slaves: out ipb_wbus);

procedure ipbus_read(
					is_last : boolean;
					signal clk: in std_logic; 
					signal cntrl_to_slaves: out ipb_wbus; 
					signal cntrl_from_slaves: in ipb_rbus; 
					add: in std_logic_vector(31 downto 0); 

					signal data: out std_logic_vector(31 downto 0));

procedure ipbus_block_read(
					is_ni : boolean;
					count : positive;
					signal clk: in std_logic; 
					signal cntrl_to_slaves: out ipb_wbus; 
					signal cntrl_from_slaves: in ipb_rbus; 
					add: in std_logic_vector(31 downto 0); 

					signal data: out std_logic_vector(31 downto 0));

procedure ipbus_write(
					is_last : boolean;
					signal clk: in std_logic; 
					signal cntrl_to_slaves: out ipb_wbus; 
					signal cntrl_from_slaves: in ipb_rbus; 
					add: in std_logic_vector(31 downto 0); 
					data: in std_logic_vector(31 downto 0));
                        
procedure ipbus_block_write(
					is_ni : boolean;
					signal clk: in std_logic; 
					signal cntrl_to_slaves: out ipb_wbus; 
					signal cntrl_from_slaves: in ipb_rbus; 
					add: in std_logic_vector(31 downto 0); 
					data: in type_ipbus_buffer);
                        
end pack_simu_transactor_lpsc;

package body pack_simu_transactor_lpsc is
-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
procedure ipbus_init(
				signal clk: in std_logic; 
				signal cntrl_to_slaves: out ipb_wbus) is
begin
	cntrl_to_slaves.ipb_addr<=(others=>'U');
	cntrl_to_slaves.ipb_wdata<=(others=>'U');
	cntrl_to_slaves.ipb_strobe<='0';
	cntrl_to_slaves.ipb_write<='0';
	wait until rising_edge(clk);
end ipbus_init;
-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
procedure ipbus_read(
					is_last : boolean;
					signal clk: in std_logic; 
					signal cntrl_to_slaves: out ipb_wbus; 
					signal cntrl_from_slaves: in ipb_rbus; 
					add: in std_logic_vector(31 downto 0); 
					signal data: out std_logic_vector(31 downto 0)) is
begin
	cntrl_to_slaves.ipb_addr<=add;
	cntrl_to_slaves.ipb_wdata<=(others=>'U');
	cntrl_to_slaves.ipb_strobe<='1';
	cntrl_to_slaves.ipb_write<='0';

	while true loop
		wait until rising_edge(clk);
		if cntrl_from_slaves.ipb_ack='1' or cntrl_from_slaves.ipb_err='1' then
			data<=cntrl_from_slaves.ipb_rdata;
			exit;
		end if;
	end loop;

	if is_last then
		cntrl_to_slaves.ipb_wdata<=(others=>'U');
		cntrl_to_slaves.ipb_strobe<='0';
		wait until rising_edge(clk);
	end if;

end ipbus_read;
-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
procedure ipbus_block_read(
					is_ni : boolean;
					count : positive;
					signal clk: in std_logic; 
					signal cntrl_to_slaves: out ipb_wbus; 
					signal cntrl_from_slaves: in ipb_rbus; 
					add: in std_logic_vector(31 downto 0); 
					signal data: out std_logic_vector(31 downto 0)) is

variable is_last : boolean;
variable cur_add : std_logic_vector(31 downto 0);
begin

for i in 0 to count-1 loop
	if is_ni then
		cur_add:=add;
	else
		cur_add:=add+conv_std_logic_vector(i,32);
	end if;
	if i=count-1 then
		is_last:=true;
	else
		is_last:=false;
	end if;
	ipbus_read(is_last,clk,cntrl_to_slaves,cntrl_from_slaves,cur_add,data);
end loop;

end ipbus_block_read;
-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
procedure ipbus_write(
					is_last : boolean;
					signal clk: in std_logic; 
					signal cntrl_to_slaves: out ipb_wbus; 
					signal cntrl_from_slaves: in ipb_rbus; 
					add: in std_logic_vector(31 downto 0); 
					data: in std_logic_vector(31 downto 0)) is
begin
	cntrl_to_slaves.ipb_addr<=add;
	cntrl_to_slaves.ipb_wdata<=data;
	cntrl_to_slaves.ipb_strobe<='1';
	cntrl_to_slaves.ipb_write<='1';

	while true loop
		wait until rising_edge(clk);
		if cntrl_from_slaves.ipb_ack='1' or cntrl_from_slaves.ipb_err='1' then
			exit;
		end if;
	end loop;

	if is_last then
		cntrl_to_slaves.ipb_wdata<=(others=>'U');
		cntrl_to_slaves.ipb_strobe<='0';
		wait until rising_edge(clk);
	end if;

end ipbus_write;

-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
procedure ipbus_block_write(
					is_ni : boolean;
					signal clk: in std_logic; 
					signal cntrl_to_slaves: out ipb_wbus; 
					signal cntrl_from_slaves: in ipb_rbus; 
					add: in std_logic_vector(31 downto 0); 
					data: in type_ipbus_buffer) is
variable is_last : boolean;
variable cur_add : std_logic_vector(31 downto 0);
variable count : natural;
begin

count:=data'length;
for i in 0 to count-1 loop
	if is_ni then
		cur_add:=add;
	else
		cur_add:=add+conv_std_logic_vector(i,32);
	end if;
	if i=count-1 then
		is_last:=true;
	else
		is_last:=false;
	end if;
	ipbus_write(is_last,clk,cntrl_to_slaves,cntrl_from_slaves,cur_add,data(i));
end loop;

end ipbus_block_write;

end pack_simu_transactor_lpsc;



