library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

use work.ipbus.all;
use work.pack_simu_transactor_lpsc.all;

entity tb_transactor_lpsc is
generic(
	-- Number of RX and TX buffers is 2**BUFWIDTH
	BUFWIDTH: natural := 4;
	-- Number of address bits within each buffer in UDP I/F
	-- Size of each buffer is 2**ADDRWIDTH
	ADDRWIDTH: natural := 9
	);
end tb_transactor_lpsc;

architecture simu of tb_transactor_lpsc is
component transactor_lpsc
generic(
	-- Number of RX and TX buffers is 2**BUFWIDTH
	BUFWIDTH: natural := 4;
	-- Number of address bits within each buffer in UDP I/F
	-- Size of each buffer is 2**ADDRWIDTH
	ADDRWIDTH: natural := 9
	);
port(
	ipb_clk: in std_logic; -- IPbus clock
	rst_ipb: in std_logic; -- IPbus clock domain sync reset

	--ipb bus
	ipb_out: out ipb_wbus; -- IPbus bus signals
	ipb_in: in ipb_rbus;

	-- status interface
	next_expected_packetid : out std_logic_vector(15 downto 0);

	-- rx buffer interface
	rxtrans_buf_wr_ptr : in std_logic_vector(BUFWIDTH downto 0);
	rxtrans_rd_ptr : out std_logic_vector(ADDRWIDTH+BUFWIDTH-1 downto 0);
	rxtrans_dat : in std_logic_vector(31 downto 0);

	-- tx buffer interface
	txtrans_buf_wr_ptr : out std_logic_vector(BUFWIDTH downto 0);
	txtrans_dat_wr_ptr : out std_logic_vector(ADDRWIDTH-1 downto 0);
	txtrans_we : out std_logic;
	txtrans_dat : out std_logic_vector(31 downto 0)
	);
end component;
component slaves
	port(
		ipb_clk: in std_logic;
		ipb_rst: in std_logic;
		ipb_in: in ipb_wbus;
		ipb_out: out ipb_rbus;
		rst_out: out std_logic;
		eth_err_ctrl: out std_logic_vector(35 downto 0);
		eth_err_stat: in std_logic_vector(47 downto 0) := X"000000000000";
		pkt_rx: in std_logic := '0';
		pkt_tx: in std_logic := '0'
	);
end component;

signal ipb_clk: std_logic; -- IPbus clock
signal rst_ipb: std_logic; -- IPbus clock domain sync reset
signal ipb_out: ipb_wbus; -- IPbus bus signals
signal ipb_in: ipb_rbus;
signal ipb_out_pack: ipb_wbus; -- IPbus bus signals
signal ipb_in_pack: ipb_rbus;

signal data_read : std_logic_vector(31 downto 0);

signal rxtrans_buf_wr_ptr : std_logic_vector(BUFWIDTH downto 0);
signal rxtrans_rd_ptr : std_logic_vector(ADDRWIDTH+BUFWIDTH-1 downto 0);
signal rxtrans_dat : std_logic_vector(31 downto 0);
signal txtrans_buf_wr_ptr : std_logic_vector(BUFWIDTH downto 0);
signal txtrans_dat_wr_ptr : std_logic_vector(ADDRWIDTH-1 downto 0);
signal txtrans_we : std_logic;
signal txtrans_dat : std_logic_vector(31 downto 0);
signal buf_sel : natural;
signal dat_sel : natural;

signal txbuf_sel : natural;
signal txdat_sel : natural;
------------------------------------------------------------------------------
-- types to support transaction buffer data
------------------------------------------------------------------------------
type data_typ is array (natural range <>) of bit_vector(31 downto 0);
type frame_typ_ary is array (natural range <>) of data_typ(0 to 40);

signal tx_buf : frame_typ_ary(0 to 8);

constant transaction_data : frame_typ_ary := (
-------------
-- Frame 0 
-------------
0          => (
	0        => x"ABBA" & '1' & to_bitvector(conv_std_logic_vector(24,15)),    --rxpacket_size
	1        => X"20_0001_F0",   --first ipbus packet header
	
	2        => X"2_000_01_1_F", --first transaction (single write)
	3        => X"0000_1000",    --first transaction base address
	4        => X"ABBA_B000",    --first transaction 1 data
	
	5        => X"2_001_03_1_F", --2nd transaction (triple write)
	6        => X"0000_1001",    --2nd transaction base address
	7        => X"ABCA_C001",    --2nd transaction data 1
	8        => X"ABCA_C002",    --2nd transaction data 2
	9        => X"ABCA_C003",    --2nd transaction data 3

	10       => X"2_002_03_0_F", --3rd transaction (triple read)
	11       => X"0000_1000",    --3rd transaction base address
	
	12       => X"2_003_04_2_F", --4th transaction (quad ni_read)
	13       => X"0000_1000",    --4th transaction base address
	
	14       => X"2_004_02_3_F", --transaction (double ni_write)
	15       => X"0000_1010",    --transaction base address
	16       => X"A000_D001",    --transaction data 1
	17       => X"A000_D002",    --transaction data 2

	18       => X"2_005_01_4_F", --transaction (RMW bit)
	19       => X"0000_1010",    --transaction base address
	20       => X"0000_F000",    --AND term
	21       => X"000A_0000",    --OR term

	22       => X"2_006_01_5_F", --transaction (RMW sum)
	23       => X"0000_1010",    --transaction base address
	24       => X"0000_00B0",    --ADDEND
	others   => X"0000_0000"),

-------------
-- Frame 1 --same as before, good number
-------------
1          => (
	0        => x"ABBA" & '1' & to_bitvector(conv_std_logic_vector(24,15)),    --rxpacket_size
	1        => X"20_0002_F0",   --first ipbus packet header
	
	2        => X"2_000_01_1_F", --first transaction (single write)
	3        => X"0000_1000",    --first transaction base address
	4        => X"ABBA_B000",    --first transaction 1 data
	
	5        => X"2_001_03_1_F", --2nd transaction (triple write)
	6        => X"0000_1001",    --2nd transaction base address
	7        => X"ABCA_C001",    --2nd transaction data 1
	8        => X"ABCA_C002",    --2nd transaction data 2
	9        => X"ABCA_C003",    --2nd transaction data 3

	10       => X"2_002_03_0_F", --3rd transaction (triple read)
	11       => X"0000_1000",    --3rd transaction base address
	
	12       => X"2_003_04_2_F", --4th transaction (quad ni_read)
	13       => X"0000_1000",    --4th transaction base address
	
	14       => X"2_004_02_3_F", --transaction (double ni_write)
	15       => X"0000_1010",    --transaction base address
	16       => X"A000_D001",    --transaction data 1
	17       => X"A000_D002",    --transaction data 2

	18       => X"2_005_01_4_F", --transaction (RMW bit)
	19       => X"0000_1010",    --transaction base address
	20       => X"0000_F000",    --AND term
	21       => X"000A_0000",    --OR term

	22       => X"2_006_01_5_F", --transaction (RMW sum)
	23       => X"0000_1010",    --transaction base address
	24       => X"0000_00B0",    --ADDEND
	others   => X"0000_0000"),

-------------
-- Frame 2 --same as before, bad packet number
-------------
2          => (
	0        => x"ABBA" & '1' & to_bitvector(conv_std_logic_vector(24,15)),    --rxpacket_size
	1        => X"20_0002_F0",   --first ipbus packet header
	
	2        => X"2_000_01_1_F", --first transaction (single write)
	3        => X"0000_1000",    --first transaction base address
	4        => X"ABBA_B000",    --first transaction 1 data
	
	5        => X"2_001_03_1_F", --2nd transaction (triple write)
	6        => X"0000_1001",    --2nd transaction base address
	7        => X"ABCA_C001",    --2nd transaction data 1
	8        => X"ABCA_C002",    --2nd transaction data 2
	9        => X"ABCA_C003",    --2nd transaction data 3

	10       => X"2_002_03_0_F", --3rd transaction (triple read)
	11       => X"0000_1000",    --3rd transaction base address
	
	12       => X"2_003_04_2_F", --4th transaction (quad ni_read)
	13       => X"0000_1000",    --4th transaction base address
	
	14       => X"2_004_02_3_F", --transaction (double ni_write)
	15       => X"0000_1010",    --transaction base address
	16       => X"A000_D001",    --transaction data 1
	17       => X"A000_D002",    --transaction data 2

	18       => X"2_005_01_4_F", --transaction (RMW bit)
	19       => X"0000_1010",    --transaction base address
	20       => X"0000_F000",    --AND term
	21       => X"000A_0000",    --OR term

	22       => X"2_006_01_5_F", --transaction (RMW sum)
	23       => X"0000_1010",    --transaction base address
	24       => X"0000_00B0",    --ADDEND
	others   => X"0000_0000"),

-------------
-- Frame 3 --same as before, 0 packet number
-------------
3          => (
	0        => x"ABBA" & '1' & to_bitvector(conv_std_logic_vector(24,15)),    --rxpacket_size
	1        => X"20_0000_F0",   --first ipbus packet header
	
	2        => X"2_000_01_1_F", --first transaction (single write)
	3        => X"0000_1000",    --first transaction base address
	4        => X"ABBA_B000",    --first transaction 1 data
	
	5        => X"2_001_03_1_F", --2nd transaction (triple write)
	6        => X"0000_1001",    --2nd transaction base address
	7        => X"ABCA_C001",    --2nd transaction data 1
	8        => X"ABCA_C002",    --2nd transaction data 2
	9        => X"ABCA_C003",    --2nd transaction data 3

	10       => X"2_002_03_0_F", --3rd transaction (triple read)
	11       => X"0000_1000",    --3rd transaction base address
	
	12       => X"2_003_04_2_F", --4th transaction (quad ni_read)
	13       => X"0000_1000",    --4th transaction base address
	
	14       => X"2_004_02_3_F", --transaction (double ni_write)
	15       => X"0000_1010",    --transaction base address
	16       => X"A000_D001",    --transaction data 1
	17       => X"A000_D002",    --transaction data 2

	18       => X"2_005_01_4_F", --transaction (RMW bit)
	19       => X"0000_1010",    --transaction base address
	20       => X"0000_F000",    --AND term
	21       => X"000A_0000",    --OR term

	22       => X"2_006_01_5_F", --transaction (RMW sum)
	23       => X"0000_1010",    --transaction base address
	24       => X"0000_00B0",    --ADDEND
	others   => X"0000_0000"),

-------------
-- Frame 4 --same as before, good packet number, timeout on 2nd and 3rd transaction
-------------
4          => (
	0        => x"ABBA" & '1' & to_bitvector(conv_std_logic_vector(24,15)),    --rxpacket_size
	1        => X"20_0003_F0",   --first ipbus packet header
	
	2        => X"2_000_01_1_F", --first transaction (single write)
	3        => X"0000_1000",    --first transaction base address
	4        => X"ABBA_B000",    --first transaction 1 data
	
	5        => X"2_001_03_1_F", --2nd transaction (triple write)
	6        => X"0000_000A",    --2nd transaction base address
	7        => X"ABCA_C001",    --2nd transaction data 1
	8        => X"ABCA_C002",    --2nd transaction data 2
	9        => X"ABCA_C003",    --2nd transaction data 3

	10       => X"2_002_03_0_F", --3rd transaction (triple read)
	11       => X"0000_000A",    --3rd transaction base address
	
	12       => X"2_003_04_2_F", --4th transaction (quad ni_read)
	13       => X"0000_1000",    --4th transaction base address
	
	14       => X"2_004_02_3_F", --transaction (double ni_write)
	15       => X"0000_1010",    --transaction base address
	16       => X"A000_D001",    --transaction data 1
	17       => X"A000_D002",    --transaction data 2

	18       => X"2_005_01_4_F", --transaction (RMW bit)
	19       => X"0000_1010",    --transaction base address
	20       => X"0000_F000",    --AND term
	21       => X"000A_0000",    --OR term

	22       => X"2_006_01_5_F", --transaction (RMW sum)
	23       => X"0000_1010",    --transaction base address
	24       => X"0000_00B0",    --ADDEND
	others   => X"0000_0000"),

-------------
-- Frame 5 --same as before, good packet number, throughput measurement
-------------
5          => (
	0        => x"ABBA" & '1' & to_bitvector(conv_std_logic_vector(5,15)),    --rxpacket_size
	1        => X"20_0004_F0",   --first ipbus packet header
	
	2        => X"2_000_FF_2_F", --first transaction (255 NI read)
	3        => X"0000_1000",    --first transaction base address
	
	4        => X"2_001_6E_2_F", --first transaction (110 NI read)
	5        => X"0000_1000",    --first transaction base address
	
	others   => X"0000_0000"),

-------------
-- Frame 6 --same as before, good packet number, throughput measurement
-------------
6          => (
	0        => x"ABBA" & '1' & to_bitvector(conv_std_logic_vector(5,15)),    --rxpacket_size
	1        => X"20_0005_F0",   --first ipbus packet header
	
	2        => X"2_000_FF_2_F", --first transaction (255 NI read)
	3        => X"0000_1000",    --first transaction base address
	
	4        => X"2_001_6E_2_F", --first transaction (110 NI read)
	5        => X"0000_1000",    --first transaction base address
	
	others   => X"0000_0000"),


7          => (
	others        => x"FFFF_FFFF") 
);

begin
p_ipb_clk : process
	begin
	ipb_clk <= '0';
	wait for 4 ns;

	ipb_clk <= '1';
	wait for 4 ns;
end process p_ipb_clk;
-------------------------------------------------------------------------------
-- DUT
-------------------------------------------------------------------------------
lbl_DUT : transactor_lpsc  generic map(
	-- Number of RX and TX buffers is 2**BUFWIDTH
	BUFWIDTH=>BUFWIDTH,
	-- Number of address bits within each buffer in UDP I/F
	-- Size of each buffer is 2**ADDRWIDTH
	ADDRWIDTH=>ADDRWIDTH
	)
port map(
	ipb_clk=>ipb_clk,
	rst_ipb=>rst_ipb,

	--ipb bus
	ipb_out=>ipb_out,
	ipb_in=>ipb_in,

	-- status interface
	next_expected_packetid=>open,

	-- rx buffer interface
	rxtrans_buf_wr_ptr=>rxtrans_buf_wr_ptr,
	rxtrans_rd_ptr=>rxtrans_rd_ptr,
	rxtrans_dat=>rxtrans_dat,

	-- tx buffer interface
	txtrans_buf_wr_ptr=>txtrans_buf_wr_ptr,
	txtrans_dat_wr_ptr=>txtrans_dat_wr_ptr,
	txtrans_we=>txtrans_we,
	txtrans_dat=>txtrans_dat
	);
-------------------------------------------------------------------------------
-- in buffer interface
-------------------------------------------------------------------------------
dat_sel<=conv_integer(rxtrans_rd_ptr(ADDRWIDTH-1 downto 0));
buf_sel<=conv_integer(rxtrans_rd_ptr(ADDRWIDTH+BUFWIDTH-1 downto ADDRWIDTH));

-- simulate BRAM behabviour
process(ipb_clk)
begin
	if rising_edge(ipb_clk) then
		rxtrans_dat<=to_stdlogicvector(transaction_data(buf_sel)(dat_sel));
	end if;
end process;
-------------------------------------------------------------------------------
-- out buffer interface
-------------------------------------------------------------------------------
txdat_sel<=conv_integer(txtrans_dat_wr_ptr) when txtrans_dat_wr_ptr<39 else 0;
txbuf_sel<=conv_integer(txtrans_buf_wr_ptr(BUFWIDTH-1 downto 0));

process(ipb_clk)
begin
	if rising_edge(ipb_clk) then
		if txtrans_we='1' then
			tx_buf(txbuf_sel)(txdat_sel)<=to_bitvector(txtrans_dat);
		end if;
	end if;
end process;
-------------------------------------------------------------------------------
-- slaves
-------------------------------------------------------------------------------
lbl_slaves : slaves port map(
		ipb_clk=>ipb_clk,
		ipb_rst=>rst_ipb,
		ipb_in=>ipb_out,
		ipb_out=>ipb_in,
		rst_out=>open,
		eth_err_ctrl=>open,
		eth_err_stat=>X"000000000000",
		pkt_rx=>'0',
		pkt_tx=>'0'
	);
-------------------------------------------------------------------------------
-- slaves
-------------------------------------------------------------------------------
lbl_slaves_pack : slaves port map(
		ipb_clk=>ipb_clk,
		ipb_rst=>rst_ipb,
		ipb_in=>ipb_out_pack,
		ipb_out=>ipb_in_pack,
		rst_out=>open,
		eth_err_ctrl=>open,
		eth_err_stat=>X"000000000000",
		pkt_rx=>'0',
		pkt_tx=>'0'
	);
-------------------------------------------------------------------------------
-- package test process
-------------------------------------------------------------------------------
pack_proc : process
variable wdata : std_logic_vector(31 downto 0);
variable ipb_add : std_logic_vector(31 downto 0);
variable rdata : std_logic_vector(31 downto 0);
begin
	ipbus_init(ipb_clk,ipb_out_pack);
	wait for 1 us;

	wdata:=x"DEAD_BEEF";
	ipbus_write(true,ipb_clk, ipb_out_pack,ipb_in_pack,x"0000_1000",wdata);
	wait for 200 ns;

	ipbus_read(true,ipb_clk, ipb_out_pack,ipb_in_pack,x"0000_1000",data_read);
-- 	data_read<=rdata;
	wait for 200 ns;

	--test write sequence
	for i in 0 to 7 loop
		wdata:=x"DEAD_BEE1"+conv_std_logic_vector(i,32);
		ipb_add:=x"0000_1000"+conv_std_logic_vector(i,32);
		if i=7 then
			ipbus_write(true,ipb_clk, ipb_out_pack,ipb_in_pack,ipb_add,wdata);
		else
			ipbus_write(false,ipb_clk, ipb_out_pack,ipb_in_pack,ipb_add,wdata);
		end if;
	end loop;
	wait for 200 ns;

	--test read sequence
	for i in 0 to 7 loop
		ipb_add:=x"0000_1000"+conv_std_logic_vector(i,32);
		if i=7 then
			ipbus_read(true,ipb_clk, ipb_out_pack,ipb_in_pack,ipb_add,data_read);
		else
			ipbus_read(false,ipb_clk, ipb_out_pack,ipb_in_pack,ipb_add,data_read);
		end if;
-- 		data_read<=rdata;
	end loop;
	wait for 200 ns;

	wait;

end process pack_proc;
-------------------------------------------------------------------------------
-- main process
-------------------------------------------------------------------------------
main : process
begin
	rst_ipb<='1';
	rxtrans_buf_wr_ptr<=(others=>'0');
	wait for 20 ns;

	rst_ipb<='0';
	wait for 100 ns;

	rxtrans_buf_wr_ptr<=conv_std_logic_vector(1,BUFWIDTH+1);
	wait for 1 us;

	rxtrans_buf_wr_ptr<=conv_std_logic_vector(5,BUFWIDTH+1);
	wait for 18 us;

	rxtrans_buf_wr_ptr<=conv_std_logic_vector(7,BUFWIDTH+1);
	wait for 18 us;

	report "fin de simulation" severity ERROR;
end process main;
end simu;
