library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_arith.ALL;
use IEEE.STD_LOGIC_unsigned.ALL;

use work.pack_ipv4_types.all;
use work.ipbus.all;

library unisim;
use unisim.VComponents.all;

entity top_ipbus_lpsc_basex is
generic (
	IN_SIMULATION     : boolean :=false;
	SIM_GTPRESET_SPEEDUP : natural :=0;
	SIMUL_DST_MAC_NUMBER  : std_logic_vector(47 downto 0):=x"AAAA_BBBB_CCCC";
	CLOCK_FREQ			: integer := 125000000;							-- freq of data_in_clk -- needed to timout cntr
	ARP_TIMEOUT			: integer := 60;									-- ARP response timeout (s)
	ARP_MAX_PKT_TMO	: integer := 5;									-- # wrong nwk pkts received before set error
	MAX_ARP_ENTRIES 	: integer := 255									-- max entries in the ARP store
	);
port(
	sysclk_p: in STD_LOGIC;
   	sysclk_n: in STD_LOGIC;

	leds: out STD_LOGIC_VECTOR(3 downto 0);

	gtp_clkp: in std_logic;
	gtp_clkn: in std_logic;
	gtp_txp: out std_logic;
	gtp_txn: out std_logic;
	gtp_rxp: in std_logic;
	gtp_rxn: in std_logic;
	sfp_los: in std_logic;
	user_sma_gpio_n : out std_logic;
	user_sma_gpio_p : out std_logic;
	
	dip_switch: in std_logic_vector(3 downto 0)
	);
end top_ipbus_lpsc_basex;

architecture RTL of top_ipbus_lpsc_basex is
component UDP_Complete
generic (
	IN_SIMULATION     : boolean :=false;
	SIMUL_DST_MAC_NUMBER  : std_logic_vector(47 downto 0):=x"AAAA_BBBB_CCCC";
	CLOCK_FREQ			: integer := 125000000;							-- freq of data_in_clk -- needed to timout cntr
	ARP_TIMEOUT			: integer := 60;									-- ARP response timeout (s)
	ARP_MAX_PKT_TMO	: integer := 5;									-- # wrong nwk pkts received before set error
	MAX_ARP_ENTRIES 	: integer := 255									-- max entries in the ARP store
	);
Port (
	-- UDP TX signals
	udp_tx_start			: in std_logic;							-- indicates req to tx UDP
	udp_txi					: in udp_tx_type;							-- UDP tx cxns
	udp_tx_result			: out std_logic_vector (1 downto 0);-- tx status (changes during transmission)
	udp_tx_data_out_ready: out std_logic;							-- indicates udp_tx is ready to take data

	-- UDP RX signals
	udp_rx_start			: out std_logic;							-- indicates receipt of udp header
	udp_rxo					: out udp_rx_type;

	-- IP RX signals
	ip_rx_hdr				: out ipv4_rx_header_type;

	-- system signals
	clk125					: in  STD_LOGIC;
	reset 					: in  STD_LOGIC;
	our_ip_address 		: in STD_LOGIC_VECTOR (31 downto 0);
	our_mac_address 		: in std_logic_vector (47 downto 0);
	control					: in udp_control_type;

	-- status signals
	arp_pkt_count			: out STD_LOGIC_VECTOR(7 downto 0);			-- count of arp pkts received
	ip_pkt_count			: out STD_LOGIC_VECTOR(7 downto 0);			-- number of IP pkts received for us
	
	-- gmii TX interface
	-- 	gmii_gtx_clk : out STD_LOGIC; -- must be generated with appropriate system
	gmii_tx_en : out STD_LOGIC;
	gmii_tx_er : out STD_LOGIC;
	gmii_txd : out STD_LOGIC_VECTOR(7 downto 0);

	-- gmii RX interface
	gmii_rx_clk: in STD_LOGIC;
   	gmii_rx_dv: in STD_LOGIC;
	gmii_rx_er: in STD_LOGIC;
	gmii_rxd : in STD_LOGIC_VECTOR(7 downto 0)
	);
end component;
component slaves
	port(
		ipb_clk: in std_logic;
		ipb_rst: in std_logic;
		ipb_in: in ipb_wbus;
		ipb_out: out ipb_rbus;
		rst_out: out std_logic;
		eth_err_ctrl: out std_logic_vector(35 downto 0);
		eth_err_stat: in std_logic_vector(47 downto 0) := X"000000000000";
		pkt_rx: in std_logic := '0';
		pkt_tx: in std_logic := '0'
	);
end component;

component ipbus_main
generic(
	-- Number of RX and TX buffers is 2**BUFWIDTH
	BUFWIDTH: natural := 4;
	-- Number of address bits within each buffer in UDP I/F
	-- Size of each buffer is 2**ADDRWIDTH --MAXIMUM is 14 bit
	ADDRWIDTH: natural := 9;
	-- UDP port for IPbus traffic in this instance of UDP I/F
	IPBUSPORT: std_logic_vector(15 DOWNTO 0) := x"C351"
	);
port(
	mac_clk: in std_logic; -- Ethernet MAC clock (125MHz)
	rst_macclk: in std_logic; -- MAC clock domain sync reset
    
	-- UDP RX signals
    udp_rx_start : in std_logic;       -- indicates receipt of udp header
    udp_rxo      : in udp_rx_type;

	-- UDP TX signals
	udp_tx_start			: out std_logic;					-- indicates req to tx UDP
	udp_txi					: out udp_tx_type;					-- UDP tx cxns
	udp_tx_result			: in std_logic_vector (1 downto 0); -- tx status (changes during transmission)
	udp_tx_data_out_ready   : in std_logic;						-- indicates udp_tx is ready to take data

	-- ipb interface
	ipb_clk: in std_logic; -- IPbus clock
	rst_ipb: in std_logic; -- IPbus clock domain sync reset
	ipb_out: out ipb_wbus; -- IPbus bus signals
	ipb_in: in ipb_rbus
	);
end component;

--UDP/IPBUS interconnection
signal udp_tx_start			:std_logic;							-- indicates req to tx UDP
signal udp_txi	: udp_tx_type;
signal udp_tx_result		: std_logic_vector (1 downto 0);-- tx status (changes during transmission)
signal udp_tx_data_out_ready: std_logic;							-- indicates udp_tx is ready to take data

signal udp_rx_start	: std_logic;							-- indicates receipt of udp header
signal udp_rxo		: udp_rx_type;

signal control			: udp_control_type;

--GMII interfacing	
signal clk125: std_logic;
signal rst_125: std_logic;
signal rst_ipb: std_logic;
signal ipb_clk: std_logic;
signal sys_rst: std_logic;
signal sys_clk: std_logic;
signal locked: std_logic;
signal onehz: std_logic;

-- interco between UDP and PHY
signal gmii_txd, gmii_rxd: std_logic_vector(7 downto 0);
signal gmii_tx_en, gmii_tx_er, gmii_rx_dv, gmii_rx_er: std_logic;
	
signal status: std_logic_vector(15 downto 0);
signal clkin, gtpclkout, gtpclkout_buf: std_logic;

signal our_ip_address 		: STD_LOGIC_VECTOR (31 downto 0);
signal our_mac_address 		: std_logic_vector (47 downto 0);

-- IPBUS connection
signal ipb_out: ipb_wbus; -- IPbus bus signals
signal ipb_in: ipb_rbus;
signal an_adv_config_vector0:std_logic_vector(15 downto 0);

begin
-------------------------------------------------------------------------------
-- UDP connection
-------------------------------------------------------------------------------
lbl_UDP : UDP_Complete generic map(
			IN_SIMULATION=>IN_SIMULATION,
			SIMUL_DST_MAC_NUMBER=>SIMUL_DST_MAC_NUMBER,

			CLOCK_FREQ		=> CLOCK_FREQ,
			ARP_TIMEOUT		=> ARP_TIMEOUT,
			ARP_MAX_PKT_TMO	=> ARP_MAX_PKT_TMO,
			MAX_ARP_ENTRIES =>MAX_ARP_ENTRIES 
			)
    port map(
			-- UDP TX signals
			udp_tx_start=>udp_tx_start,
			udp_txi=>udp_txi,
			udp_tx_result=>udp_tx_result,
			udp_tx_data_out_ready=>udp_tx_data_out_ready,
			
			-- UDP RX signals
			udp_rx_start=>udp_rx_start,
			udp_rxo=>udp_rxo,
			
			-- IP RX signals
			ip_rx_hdr=>open,

			-- system signals
			clk125=>clk125,
			reset=>rst_125,
			our_ip_address=>our_ip_address,
			our_mac_address=>our_mac_address,
			control=>control,

			-- status signals
			arp_pkt_count=>open,
			ip_pkt_count=>open,
			
			-- gmii TX interface
			-- 	gmii_gtx_clk : out STD_LOGIC; -- must be generated with appropriate system
			gmii_txd => gmii_txd,
			gmii_tx_en => gmii_tx_en,
			gmii_tx_er => gmii_tx_er,

			-- gmii RX interface
			gmii_rx_clk=>clk125,
			gmii_rxd =>gmii_rxd,
			gmii_rx_dv=>gmii_rx_dv,
			gmii_rx_er=>gmii_rx_er
			);
-- our_mac_address<= X"020ddba115" & X"0" & X"0"; -- Careful here, arbitrary addresses do not always work
our_mac_address<= X"020ddba115" & X"3" & X"0"; -- Careful here, arbitrary addresses do not always work
our_ip_address<= X"c0a864" & X"10"; -- 192.168.100.16

control.ip_controls.arp_controls.clear_cache<='0';
-------------------------------------------------------------------------------
--	IPBUS block
-------------------------------------------------------------------------------
lbl_ipbus : ipbus_main generic map(
	-- Number of RX and TX buffers is 2**BUFWIDTH
	BUFWIDTH=>4,
	-- Number of address bits within each buffer in UDP I/F
	-- Size of each buffer is 2**ADDRWIDTH --MAXIMUM is 14 bit
	ADDRWIDTH=>9,
	-- UDP port for IPbus traffic in this instance of UDP I/F
	IPBUSPORT=>x"C351"
	)
port map(
	mac_clk=>clk125,
	rst_macclk=>rst_125,
    
	-- UDP RX signals
	udp_rx_start=>udp_rx_start,
	udp_rxo=>udp_rxo,
	
	-- UDP TX signals
	udp_tx_start=>udp_tx_start,
	udp_txi=>udp_txi,
	udp_tx_result=>udp_tx_result,
	udp_tx_data_out_ready=>udp_tx_data_out_ready,

	-- ipb interface
	ipb_clk=>ipb_clk,
	rst_ipb=>rst_ipb,
	ipb_out=>ipb_out,
	ipb_in=>ipb_in
	);
-------------------------------------------------------------------------------
--	slaves block
-------------------------------------------------------------------------------
lbl_slaves : slaves port map(
		ipb_clk=>ipb_clk,
		ipb_rst=>rst_ipb,
		ipb_in=>ipb_out,
		ipb_out=>ipb_in,
		rst_out=>open,
		eth_err_ctrl=>open,
		eth_err_stat=>X"000000000000",
		pkt_rx=>'0',
		pkt_tx=>'0'
	);
-------------------------------------------------------------------------------
--	DCM clock generation for internal bus, ethernet
-------------------------------------------------------------------------------
clocks: entity work.clocks_s6_basex generic map (fast_sim=>IN_SIMULATION)
	port map(
	sysclk_p => sysclk_p,
	sysclk_n => sysclk_n,		
	clki_125 => clk125,
	clko_ipb => ipb_clk,
	sysclko => sys_clk,
	locked => locked,
	nuke => sys_rst,
	rsto_125 => rst_125,
	rsto_ipb => rst_ipb,
	onehz => onehz
	);

sys_rst<='0';
-------------------------------------------------------------------------------
-- GTP PHY interface interface
-------------------------------------------------------------------------------
ibuf0: IBUFDS port map(i =>gtp_clkp,ib =>gtp_clkn,o => clkin);

bufio0: BUFIO2 generic map(divide => 1,divide_bypass => true)
	port map(
		divclk => gtpclkout_buf,
		i => gtpclkout,
		ioclk => open,
		serdesstrobe => open
		);

bufg0: BUFG port map(i => gtpclkout_buf,o => clk125);

phy: entity work.gig_eth_pcs_pma_v11_3_block generic map (SIM_GTPRESET_SPEEDUP)
			port map(
			gtpclkout => gtpclkout,
-- 			gtpreset0 => '0',  -- rely upon powerup reset for now
			gtpreset0 => rst_125, 
			gmii_txd0 => gmii_txd,
			gmii_tx_en0 => gmii_tx_en,
			gmii_tx_er0 => gmii_tx_er,
			gmii_rxd0 => gmii_rxd,
			gmii_rx_dv0 => gmii_rx_dv,
			gmii_rx_er0 => gmii_rx_er,
			gmii_isolate0 => open,
-- 			configuration_vector0 => "00000", -- disable autoneg
			configuration_vector0 => "10000", -- allow autoneg
			link_timer_value0=>conv_std_logic_vector(4000,9),  -- Programmable Auto-Negotiation Link Timer Control
			an_interrupt0=>open,
			an_adv_config_vector0=>an_adv_config_vector0,
			an_restart_config0=>'1',
			status_vector0 => status,
			test=>user_sma_gpio_n,
			reset0 => rst_125,
			signal_detect0 => '1',
			clkin => clkin,
			userclk2 => clk125,
			txp0 => gtp_txp,
			txn0 => gtp_txn,
			rxp0 => gtp_rxp,
			rxn0 => gtp_rxn,
			txp1 => open,
			txn1 => open,
			rxp1 => '0',
			rxn1 => '0'
		);
user_sma_gpio_p<='0';

an_adv_config_vector0(0)<='0'; -- reserved
an_adv_config_vector0(4 downto 1)<="0000"; -- reserved
an_adv_config_vector0(5)<='1'; -- full duplex mode
an_adv_config_vector0(6)<='0'; -- reserved
an_adv_config_vector0(8 downto 7)<="00"; -- no pause
an_adv_config_vector0(9)<='0'; -- reserved
an_adv_config_vector0(11 downto 10)<="00"; -- reserved
an_adv_config_vector0(13 downto 12)<="00"; -- no error (remote fault)
an_adv_config_vector0(14)<='0'; -- reserved
an_adv_config_vector0(15)<='0'; -- reserved

end RTL;
