-------------------------------------------------------------------------------
-- manage the transaction from the data contained in the rx buffers
-- in rxbuff at add=0, the last write pointer is stored (size-1), and the src port
-- IPBUS packet starts at add=1
--
-- Assumption : the transactor_lpsc cannot be sollicited if no 
-- txtrans buffer available (ensured by IPBUS driver software)
--
-- the rx transation buffer will be arranged as follow
-- address 0x000 : udp containt length and endianess, and dst port
-- address 0x001 : client IP address
-- address 0x002 and above ipbus packet in adequat endianess

-- the tx transation buffer will be arranged as follow (note inversion from rx)
-- address 0x000 : client IP address
-- address 0x001 : udp containt length and endianess, and dst port
-- address 0x002 and above ipbus packet in adequat endianess
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

use work.ipbus.all;

entity transactor_lpsc is
generic(
	-- Number of RX and TX buffers is 2**BUFWIDTH
	BUFWIDTH: natural := 4;
	-- Number of address bits within each buffer in UDP I/F
	-- Size of each buffer is 2**ADDRWIDTH --MAXIMUM is 14 bit
	ADDRWIDTH: natural := 9
	);
port(
	ipb_clk: in std_logic; -- IPbus clock
	rst_ipb: in std_logic; -- IPbus clock domain sync reset

	--ipb bus
	ipb_out: out ipb_wbus; -- IPbus bus signals
	ipb_in: in ipb_rbus;

	-- status interface
	next_expected_packetid : out std_logic_vector(15 downto 0);

	-- rx buffer interface
	rxtrans_buf_wr_ptr : in std_logic_vector(BUFWIDTH downto 0);
	rxtrans_rd_ptr : out std_logic_vector(ADDRWIDTH+BUFWIDTH-1 downto 0);
	rxtrans_dat : in std_logic_vector(31 downto 0);

	-- tx buffer interface
	txtrans_buf_wr_ptr : out std_logic_vector(BUFWIDTH downto 0);
	txtrans_dat_wr_ptr : out std_logic_vector(ADDRWIDTH-1 downto 0);
	txtrans_we : out std_logic;
	txtrans_dat : out std_logic_vector(31 downto 0)
	);
end transactor_lpsc;

architecture RTL of transactor_lpsc is
constant TIMEOUT: integer := 255;

constant TRANS_RD: std_logic_vector(3 downto 0) := X"0";
constant TRANS_WR: std_logic_vector(3 downto 0) := X"1";
constant TRANS_NIRD: std_logic_vector(3 downto 0) := X"2";
constant TRANS_NIWR: std_logic_vector(3 downto 0) := X"3";
constant TRANS_RMWB: std_logic_vector(3 downto 0) := X"4";
constant TRANS_RMWS: std_logic_vector(3 downto 0) := X"5";
constant TRANS_RD_CFG: std_logic_vector(3 downto 0) := X"6";
constant TRANS_WR_CFG: std_logic_vector(3 downto 0) := X"7";

constant protocol_version: std_logic_vector(3 downto 0) := X"2";

type def_state_transaction is (IDLE,DO_READ_INIT,PKT_PREPAR,PKT_IP,PKT_PAR,PKT_HDR,TRANS_HDR,GET_BADD,
		EXE_TRANS,GET_PAR1,GET_PAR2,EXE_RMW_TRANS,WR_TRANS_HDR,FINISHING,FINISHING2);
signal state_transaction : def_state_transaction;
constant wait_fwft_val : natural:=3;
signal cpt_wait_fwft : natural range 0 to wait_fwft_val-1;

signal cur_rxtrans_buf_ptr : std_logic_vector(BUFWIDTH downto 0);
signal cur_rxtrans_dat_ptr : std_logic_vector(ADDRWIDTH-1 downto 0);
signal rxpacket_read : std_logic;

signal is_LE : std_logic;
signal prev_txtrans_buf_wr_ptr : std_logic_vector(BUFWIDTH downto 0);
signal prev_rxpacket_size : std_logic_vector(ADDRWIDTH-1 downto 0);
signal rxpacket_size : std_logic_vector(ADDRWIDTH-1 downto 0);
signal txpacket_size : std_logic_vector(ADDRWIDTH-1 downto 0);
signal expected_packetid : std_logic_vector(15 downto 0);
signal transaction_hdr_tx_ptr : std_logic_vector(ADDRWIDTH-1 downto 0);
signal transaction_dat_wr_ptr : std_logic_vector(ADDRWIDTH-1 downto 0);
signal word_to_go : std_logic_vector(7 downto 0);
signal transaction_size : std_logic_vector(7 downto 0);
signal transaction_type : std_logic_vector(3 downto 0);
signal transaction_ID : std_logic_vector(11 downto 0);
constant timeout_val : std_logic_vector(7 downto 0):=conv_std_logic_vector(255,8);
signal timer : std_logic_vector(7 downto 0);
-- signal rcv_packetid : std_logic_vector(15 downto 0);
signal inc_expected_pkt : std_logic;
signal rxtrans_dat_1ck : std_logic_vector(31 downto 0);
signal rxtrans_dat_fwft : std_logic_vector(31 downto 0);
signal rxtrans_dat_LE : std_logic_vector(31 downto 0);
signal txtrans_dat_LE : std_logic_vector(31 downto 0);
signal tx_info_code : std_logic_vector(3 downto 0);

signal ipb_addr : std_logic_vector(31 downto 0);
signal param1 : std_logic_vector(31 downto 0);
signal param2 : std_logic_vector(31 downto 0);
signal ipb_write : std_logic;
signal do_inc_add : std_logic;

signal dst_port : std_logic_vector(15 downto 0);
signal dst_ip : std_logic_vector(31 downto 0);

signal rd_next : std_logic;
signal rd_next_1ck : std_logic;
signal dsel : std_logic;
signal rd_init : std_logic;
signal rd_init_1ck : std_logic;

begin
next_expected_packetid<=expected_packetid;
-------------------------------------------------------------------------------
-- FWFT RAM
-------------------------------------------------------------------------------
fwft : process(rst_ipb,ipb_clk)
begin
	if rst_ipb='1' then
		cur_rxtrans_dat_ptr<=(others=>'0');
		rxtrans_dat_1ck<=(others=>'0');
		rd_init_1ck<='0';
		rd_next_1ck<='0';
		dsel<='0';
		rxpacket_read<='0';
	elsif rising_edge(ipb_clk) then
		if rd_init='1' then
			cur_rxtrans_dat_ptr<=(others=>'0');
		elsif rd_next='1' or rd_init_1ck='1' then
			cur_rxtrans_dat_ptr<=cur_rxtrans_dat_ptr+1;
		end if;
		
		rd_init_1ck<=rd_init;
		rd_next_1ck<=rd_next;
		dsel <= rd_next and rd_next_1ck;
			
		if rd_init = '1' or rd_init_1ck = '1' or rd_next = '1' or rd_next_1ck = '1' then
			rxtrans_dat_1ck<=rxtrans_dat;
		end if;
		
		if rd_init='1' then
			rxpacket_read<='0';
		elsif rd_next = '1' then
			if cur_rxtrans_dat_ptr=rxpacket_size then
				rxpacket_read<='1';
			end if;
		end if;
			
	end if;
end process fwft;
rxtrans_dat_fwft <= rxtrans_dat_1ck when dsel = '0' else rxtrans_dat;
-------------------------------------------------------------------------------
-- Transaction decoder
-------------------------------------------------------------------------------
rxtrans_rd_ptr<=cur_rxtrans_buf_ptr(BUFWIDTH-1 downto 0) & cur_rxtrans_dat_ptr;

rxtrans_dat_LE<=rxtrans_dat_fwft when is_LE='1' else 
				rxtrans_dat_fwft(7 downto 0) & rxtrans_dat_fwft(15 downto 8) & rxtrans_dat_fwft(23 downto 16) &  rxtrans_dat_fwft(31 downto 24);

txtrans_dat<=txtrans_dat_LE;
-- txtrans_dat<=txtrans_dat_LE when is_LE='1' else
-- 				txtrans_dat_LE(7 downto 0) & txtrans_dat_LE(15 downto 8) & txtrans_dat_LE(23 downto 16) &  txtrans_dat_LE(31 downto 24);

txtrans_dat_wr_ptr<=transaction_dat_wr_ptr;

ipb_out.ipb_addr<=ipb_addr;
ipb_out.ipb_write<=ipb_write;

proc_rx_mgr : process(rst_ipb,ipb_clk)
begin
	if rst_ipb='1' then
		cur_rxtrans_buf_ptr<=(others=>'0');
		rd_init<='0';
		rd_next<='0';
		dst_port<=(others=>'0');
		dst_ip<=(others=>'0');
		prev_txtrans_buf_wr_ptr<=(others=>'0');
		txtrans_buf_wr_ptr<=(others=>'0');
		is_LE<='1';
		inc_expected_pkt<='0';
		txtrans_we<='0';
		do_inc_add<='0';
		expected_packetid<=conv_std_logic_vector(1,16);
		ipb_addr<=(others=>'0');
		txpacket_size<=conv_std_logic_vector(0,ADDRWIDTH);
		rxpacket_size<=(others=>'1');
		prev_rxpacket_size<=(others=>'1');

		transaction_size<=(others=>'0');
		transaction_ID<=(others=>'0');
		param1<=(others=>'0');
		param2<=(others=>'0');
		timer<=(others=>'0');
		ipb_out.ipb_wdata<=(others=>'0');
		ipb_out.ipb_strobe<='0';
		ipb_write<='0';
		transaction_hdr_tx_ptr<=conv_std_logic_vector(3,ADDRWIDTH);
		transaction_dat_wr_ptr<=conv_std_logic_vector(1,ADDRWIDTH);
		tx_info_code<=x"0"; -- init with good transaction
		transaction_type<=(others=>'0');
		txtrans_dat_LE<=(others=>'0');
		word_to_go<=(others=>'0');
-- 		rcv_packetid<=(others=>'0');
		cpt_wait_fwft<=0;

		state_transaction<=IDLE;

	elsif rising_edge(ipb_clk) then
		txtrans_buf_wr_ptr<=prev_txtrans_buf_wr_ptr;

		case state_transaction is 
			when IDLE =>
				txtrans_we<='0';
				do_inc_add<='0';
				inc_expected_pkt<='0';
				cpt_wait_fwft<=0;
				txpacket_size<=conv_std_logic_vector(0,ADDRWIDTH);
				rd_init<='1';
				rxpacket_size<=(others=>'1'); -- to avoid early end of packet detection
				if (cur_rxtrans_buf_ptr/=rxtrans_buf_wr_ptr) then
					txtrans_we<='1';
					transaction_dat_wr_ptr<=conv_std_logic_vector(0,ADDRWIDTH);
					txtrans_dat_LE<= (others=>'0'); --to avoid resend when packet not finished (size 0)
					state_transaction<=DO_READ_INIT;
				end if;

			when DO_READ_INIT =>
				txtrans_we<='0';
				rd_init<='1';
				state_transaction<=PKT_PREPAR;

			when PKT_PREPAR =>
				rd_init<='0';
				if cpt_wait_fwft=wait_fwft_val-1 then
					rd_next<='1';
					state_transaction<=PKT_PAR;
				else
					cpt_wait_fwft<=cpt_wait_fwft+1;
				end if;

			when PKT_PAR =>
				is_LE<=rxtrans_dat_fwft(15);
				dst_port<=rxtrans_dat_fwft(31 downto 16);
				prev_rxpacket_size<=rxtrans_dat_fwft(ADDRWIDTH-1 downto 0);
				state_transaction<=PKT_IP;
				
			when PKT_IP =>
				rxpacket_size<=prev_rxpacket_size+1; -- for IP storage
				dst_ip<=rxtrans_dat_fwft;
				state_transaction<=PKT_HDR;

			when PKT_HDR =>
-- 				rcv_packetid<=rxtrans_dat_LE(23 downto 8);
				transaction_hdr_tx_ptr<=conv_std_logic_vector(3,ADDRWIDTH);
				transaction_size<=conv_std_logic_vector(1,8);
				if (expected_packetid/=rxtrans_dat_LE(23 downto 8)) then
					if rxtrans_dat_LE(23 downto 8)=0 then -- always accept
						txtrans_we<='1';
						txpacket_size<=txpacket_size+1;
						transaction_dat_wr_ptr<=conv_std_logic_vector(2,ADDRWIDTH);
						txtrans_dat_LE<= X"20" & x"0000" & x"F0";
						state_transaction<=TRANS_HDR;
					else --reject packet
						cur_rxtrans_buf_ptr<=cur_rxtrans_buf_ptr+1;
						rd_next<='0';
						state_transaction<=IDLE;
					end if;
				else -- accept and increment next packet header
					txtrans_we<='1';
					txpacket_size<=txpacket_size+1;
					txtrans_dat_LE<= X"20" & rxtrans_dat_LE(23 downto 8) & x"F0";
					transaction_dat_wr_ptr<=conv_std_logic_vector(2,ADDRWIDTH);
					inc_expected_pkt<='1';
					state_transaction<=TRANS_HDR;
				end if;

			when TRANS_HDR =>
				timer<=timeout_val;
				txtrans_we<='0';
				rd_next<='1';

				if rxtrans_dat_LE(31 downto 28)/=protocol_version or rxtrans_dat_LE(3 downto 0)/=x"F" then
					tx_info_code<=x"1";
					transaction_size<=rxtrans_dat_LE(15 downto 8); --necessary to debug in case of failing transaction
					transaction_ID<=rxtrans_dat_LE(27 downto 16);
					transaction_type<=rxtrans_dat_LE(7 downto 4);
					state_transaction<=WR_TRANS_HDR;
				else
					tx_info_code<=x"0"; -- init with good transaction
					word_to_go<=rxtrans_dat_LE(15 downto 8);
					transaction_size<=rxtrans_dat_LE(15 downto 8);
					transaction_ID<=rxtrans_dat_LE(27 downto 16);
					transaction_type<=rxtrans_dat_LE(7 downto 4);
					state_transaction<=GET_BADD;
				end if;

			when GET_BADD =>
				ipb_addr<=rxtrans_dat_LE;
				transaction_dat_wr_ptr<=transaction_hdr_tx_ptr;
				word_to_go<=word_to_go-1;
				rd_next<='0';

				case transaction_type is
					when TRANS_RD =>
						ipb_out.ipb_strobe<='1';
						ipb_write<='0';
						do_inc_add<='1';
						state_transaction<=EXE_TRANS;
					when TRANS_WR =>
						do_inc_add<='1';
						state_transaction<=GET_PAR1;
					when TRANS_NIRD =>
						ipb_write<='0';
						ipb_out.ipb_strobe<='1';
						do_inc_add<='0';
						state_transaction<=EXE_TRANS;
					when TRANS_NIWR =>
						do_inc_add<='0';
						state_transaction<=GET_PAR1;
					when TRANS_RMWB =>
						rd_next<='1';
						ipb_write<='0';
						do_inc_add<='0';
						state_transaction<=GET_PAR1;
					when TRANS_RMWS =>
						ipb_write<='0';
						do_inc_add<='0';
						state_transaction<=GET_PAR1;
					when others => -- problem with this packet
						state_transaction<=WR_TRANS_HDR;
				end case;

			when GET_PAR1 =>
				param1<=rxtrans_dat_LE;
				case transaction_type is
					when TRANS_RMWB =>
						rd_next<='1';
						state_transaction<=GET_PAR2;
					when TRANS_RMWS =>
						rd_next<='0';
						state_transaction<=EXE_RMW_TRANS;
					when others => --necessarly (one of the write transaction)
						rd_next<='1';
						ipb_out.ipb_strobe<='1';
						ipb_write<='1';
						ipb_out.ipb_wdata<=rxtrans_dat_LE;
						state_transaction<=EXE_TRANS;
				end case;

			when GET_PAR2 =>
				param2<=rxtrans_dat_LE;
				rd_next<='0';
				ipb_out.ipb_strobe<='1';
				state_transaction<=EXE_RMW_TRANS;

			when EXE_RMW_TRANS =>
				if ipb_in.ipb_err='1' then
					tx_info_code<="010" & ipb_write; --0x4 or 0x5
					state_transaction<=WR_TRANS_HDR;
				elsif timer=0 then --timeout
					tx_info_code<="011" & ipb_write; --0x6 or 0x7
					state_transaction<=WR_TRANS_HDR;
				elsif ipb_in.ipb_ack='1' then
					timer<=timeout_val;
-- 					ipb_out.ipb_strobe<='0';
					if (transaction_type=TRANS_RMWB) then
						ipb_out.ipb_wdata<=(ipb_in.ipb_rdata and param1) or param2;
					else
						ipb_out.ipb_wdata<=ipb_in.ipb_rdata + param1;
					end if;
					
					txtrans_we<='1';
					txpacket_size<=txpacket_size+1;
					txtrans_dat_LE<=ipb_in.ipb_rdata;
					transaction_dat_wr_ptr<=transaction_dat_wr_ptr+1;
					ipb_write<='1';
					state_transaction<=EXE_TRANS;
				else
					rd_next<='0';
					txtrans_we<='0';
					ipb_out.ipb_strobe<='1';
					timer<=timer-1;
				end if;

			when EXE_TRANS =>
				if ipb_in.ipb_err='1' then
					tx_info_code<="010" & ipb_write; --0x4 or 0x5
				end if;
				if timer=0 then --timeout
					tx_info_code<="011" & ipb_write; --0x6 or 0x7
				end if;
				
				if (ipb_in.ipb_ack='1' or timer=0 or ipb_in.ipb_err='1') then
					timer<=timeout_val;
					ipb_out.ipb_wdata<=rxtrans_dat_LE; --unconditionnal, doesn't matter
					if do_inc_add='1' then
						ipb_addr<=ipb_addr+1;
					end if;

					if ipb_write='0' then
						txtrans_we<='1';
						txpacket_size<=txpacket_size+1;
						txtrans_dat_LE<=ipb_in.ipb_rdata;
						transaction_dat_wr_ptr<=transaction_dat_wr_ptr+1;
					end if;
					
					if word_to_go/=0 then
						if ipb_write='1' and rxpacket_read='0' then
							rd_next<='1';
						else
							rd_next<='0';
						end if;
						word_to_go<=word_to_go-1;
					else
						rd_next<='0';
						ipb_out.ipb_strobe<='0';
						state_transaction<=WR_TRANS_HDR;
					end if;
				else
					rd_next<='0';
					txtrans_we<='0';
					ipb_out.ipb_strobe<='1';
					timer<=timer-1;
				end if;

			when WR_TRANS_HDR =>
				txtrans_we<='1';
				txpacket_size<=txpacket_size+1;
				txtrans_dat_LE<= protocol_version & transaction_id & transaction_size & transaction_type & tx_info_code;
				transaction_dat_wr_ptr<=transaction_hdr_tx_ptr;
				if transaction_type/=TRANS_WR and transaction_type/=TRANS_NIWR then
					transaction_hdr_tx_ptr<=transaction_hdr_tx_ptr+1+transaction_size;
				else
					transaction_hdr_tx_ptr<=transaction_hdr_tx_ptr+1;
				end if;

				if rxpacket_read='1' then
					rd_next<='1';
					state_transaction<=FINISHING;
				else
					rd_next<='1';
					state_transaction<=TRANS_HDR;
				end if;
			
			when FINISHING =>
				rd_next<='0';
				txtrans_we<='1';
				transaction_dat_wr_ptr<=conv_std_logic_vector(1,ADDRWIDTH);
				txtrans_dat_LE(31 downto 16)<=dst_port;
				txtrans_dat_LE(15)<=is_LE;
				txtrans_dat_LE(14 downto 0)<=EXT(txpacket_size,15);
				state_transaction<=FINISHING2;
			
			when FINISHING2 =>
				rd_next<='0';
				txtrans_we<='1';
				transaction_dat_wr_ptr<=conv_std_logic_vector(0,ADDRWIDTH);
				cur_rxtrans_buf_ptr<=cur_rxtrans_buf_ptr+1;
				txtrans_dat_LE(31 downto 0)<=dst_ip;
				prev_txtrans_buf_wr_ptr<=prev_txtrans_buf_wr_ptr+1;
				if inc_expected_pkt='1' then
					if expected_packetid=x"FFFF" then
						expected_packetid<=conv_std_logic_vector(1,16);
					else
						expected_packetid<=expected_packetid+1;
					end if;
				end if;
				state_transaction<=IDLE;
		end case;
	end if;
end process proc_rx_mgr;
end RTL;
