--------------------------------------------------
-- ntp_clock_filter
-- 
-- takes input tuples and find smallest delta
-- tuples with smallest delta is then used unless tuples already used
-- 
-- no jitter comparison for spike suppression
-- epsilon not computed for clock selection (several servers)
-- 
-- Jitter is computed (not RMS as in rfc 5905)
-- SGATE is set to 2 (jitter is overestimated wrt to RMS jitter)
-- allows to use a simple shift for multiplier


-- O. Bourrion May 2015 (addition of jitter selection july 2015)
--------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity ntp_clock_filter is 
port(
	refclk: in std_logic; -- Stable reference clock
	rst: in std_logic; -- rst
    
	update_tuple: in std_logic;
	is_valid: in std_logic;
	theta : in std_logic_vector(63 downto 0);
	delta : in std_logic_vector(63 downto 0);
	epoch  : in std_logic_vector(63 downto 0);

	update_local_time: out std_logic;
	allow_freq_adj: out std_logic;
	new_delta : out std_logic_vector(63 downto 0);
	new_theta : out std_logic_vector(63 downto 0);
	n : out std_logic_vector(5 downto 0); -- dimensionned to 8 full refresh of shift register of 8 values
	reference_time : out std_logic_vector(63 downto 0) -- last time when system was updated
	);
end ntp_clock_filter;

architecture RTL of ntp_clock_filter is
component ntp_addsub is 
generic(
	is_adder : boolean :=true;
	size : natural :=64 -- must be even number here
	);
port(
	clk : in std_logic;

	op_1 : in std_logic_vector(size-1 downto 0);
	op_2 : in std_logic_vector(size-1 downto 0);

	result : out std_logic_vector(size-1 downto 0)
	);
end component ntp_addsub;

type tuple is record
	theta : std_logic_vector(63 downto 0);
	delta : std_logic_vector(63 downto 0);
	epoch  : std_logic_vector(63 downto 0);
	is_valid: std_logic;
end record;

type def_tuple_mem is array(0 to 0) of tuple;
signal successive_tuples : def_tuple_mem;

signal update_tuple_1ck : std_logic;
signal check_theta : std_logic;

begin
-------------------------------------------------------------------------------
-- incoming tuple saving
-------------------------------------------------------------------------------
process(refclk)
begin
	if rising_edge(refclk) then
		if rst='1' then
			update_tuple_1ck<='0';
			update_local_time<='0';
			allow_freq_adj<='0';

		else
			if update_tuple='1' then
				successive_tuples(0).theta<=theta;
				successive_tuples(0).delta<=delta;
				successive_tuples(0).epoch<=epoch;
				successive_tuples(0).is_valid<=is_valid;
			end if;
			update_tuple_1ck<=update_tuple;

			if update_tuple_1ck='1' and successive_tuples(0).is_valid='1' then
				check_theta<='1';
			else
				check_theta<='0';
			end if;

			if check_theta='1' then
				if successive_tuples(0).theta(63 downto 32)/=X"0000_0000" and successive_tuples(0).theta(63 downto 32)/=X"FFFF_FFFF" then --below 1s offset
					allow_freq_adj<='0';
				else
					allow_freq_adj<='1';
				end if;
				new_theta<=successive_tuples(0).theta;
				new_delta<=successive_tuples(0).delta;
				reference_time<=successive_tuples(0).epoch;

				update_local_time<='1';
			else
				update_local_time<='0';
			end if;
		end if;
	end if;
end process;

n<=(others=>'0');
end RTL;
