--------------------------------------------------
-- ntp_clock_filter
-- 
-- takes input tuples and find smallest delta
-- tuples with smallest delta is then used unless tuples already used
-- 
-- no jitter comparison for spike suppression
-- epsilon not computed for clock selection (several servers)
-- 
-- Jitter is computed (not RMS as in rfc 5905)
-- SGATE is set to 2 (jitter is overestimated wrt to RMS jitter)
-- allows to use a simple shift for multiplier


-- O. Bourrion May 2015 (addition of jitter selection july 2015)
--------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity ntp_clock_filter is 
port(
	refclk: in std_logic; -- Stable reference clock
	rst: in std_logic; -- rst
    
	in_init0: in std_logic;
	in_init1: in std_logic;

	update_tuple: in std_logic;
	is_valid: in std_logic;
	theta : in std_logic_vector(63 downto 0);
	delta : in std_logic_vector(63 downto 0);
	epoch  : in std_logic_vector(63 downto 0);

	update_local_time: out std_logic;
	new_delta : out std_logic_vector(63 downto 0);
	new_theta : out std_logic_vector(63 downto 0);
	n : out std_logic_vector(5 downto 0); -- dimensionned to 8 full refresh of shift register of 8 values
	reference_time : out std_logic_vector(63 downto 0) -- last time when system was updated
	);
end ntp_clock_filter;

architecture RTL of ntp_clock_filter is
component ntp_addsub is 
generic(
	is_adder : boolean :=true;
	size : natural :=64 -- must be even number here
	);
port(
	clk : in std_logic;

	op_1 : in std_logic_vector(size-1 downto 0);
	op_2 : in std_logic_vector(size-1 downto 0);

	result : out std_logic_vector(size-1 downto 0)
	);
end component ntp_addsub;

type tuple is record
	theta : std_logic_vector(63 downto 0);
	delta : std_logic_vector(63 downto 0);
	epoch  : std_logic_vector(63 downto 0);
	is_valid: std_logic;
end record;

type def_tuple_mem is array(0 to 7) of tuple;
signal successive_tuples : def_tuple_mem;

constant comp_dly : natural:=3;
signal update_cnt : natural range 0 to comp_dly-1;

signal ptr : natural range 0 to 7;
signal cnt_since_last_upd : std_logic_vector(n'length-1 downto 0);
signal is_twice_poll_interval : std_logic;

signal best_delta_ptr : natural range 0 to 7;
signal current_delta : std_logic_vector(63 downto 0);
signal candidate_delta : std_logic_vector(63 downto 0);
signal last_epoch_used : std_logic_vector(63 downto 0);
signal last_theta_used : std_logic_vector(63 downto 0);
signal sub_op_1 : std_logic_vector(63 downto 0);
signal sub_op_2 : std_logic_vector(63 downto 0);
signal sub_result : std_logic_vector(63 downto 0);

signal update_tuple_1ck : std_logic;
signal MSB_are_equals : std_logic;
signal LSB_are_equals : std_logic;
signal longs_are_equals : std_logic;
signal MSB_smaller : std_logic;
signal LSB_smaller : std_logic;
signal new_is_smaller : std_logic;

type def_state_sort_tuple is (idle, update_comp_value,wait_result,
		update_epoch_value, wait_epoch_result,calc_diff_i,
		put_abs,accumulate_jitter,compute_theta_diff,compare_jitter);
signal state_sort_tuple : def_state_sort_tuple;

signal jitter_acc : std_logic_vector(34 downto 0);
signal jitter : std_logic_vector(32 downto 0);
signal diff_i : std_logic_vector(31 downto 0);
constant MAX_DISP : std_logic_vector(31 downto 0):=X"7FFF_FFFF"; -- 0.5sec

begin
-------------------------------------------------------------------------------
-- incoming tuple saving
-------------------------------------------------------------------------------
process(refclk)
begin
	if rising_edge(refclk) then
		if rst='1' then
			for i in 0 to 7 loop
				successive_tuples(i).is_valid<='0';
			end loop;
			update_tuple_1ck<='0';

		elsif update_tuple='1' then
			successive_tuples(0).theta<=theta;
			successive_tuples(0).delta<=delta;
			successive_tuples(0).epoch<=epoch;
			successive_tuples(0).is_valid<=is_valid;

			for i in 1 to 7 loop
				successive_tuples(i)<=successive_tuples(i-1);
			end loop;
		end if;
		update_tuple_1ck<=update_tuple;
	end if;
end process;

-------------------------------------------------------------------------------
-- 64 bit signed comparator
-------------------------------------------------------------------------------
process(refclk)
begin
	if rising_edge(refclk) then
		if candidate_delta(63 downto 32)=current_delta(63 downto 32) then
			MSB_are_equals<='1';
		else
			MSB_are_equals<='0';
		end if;

		if candidate_delta(31 downto 0)=current_delta(31 downto 0) then
			LSB_are_equals<='1';
		else
			LSB_are_equals<='0';
		end if;

		if candidate_delta(63 downto 32)<current_delta(63 downto 32) then
			MSB_smaller<='1';
		else
			MSB_smaller<='0';
		end if;

		if candidate_delta(31 downto 0)<current_delta(31 downto 0) then
			LSB_smaller<='1';
		else
			LSB_smaller<='0';
		end if;

		if MSB_are_equals='1' then -- MSB are the same (check LSB to decide)
			if LSB_are_equals='1' then
				new_is_smaller<='0';
			else
				if LSB_smaller='1' then
					new_is_smaller<='1';
				else
					new_is_smaller<='0';
				end if;
			end if;
		else --MSB are differents, check which is smaller
			if MSB_smaller='1' then
				new_is_smaller<='1';
			else
				new_is_smaller<='0';
			end if;
		end if;

-- 		if MSB_smaller='1' and MSB_are_equals='0' then
-- 			new_is_smaller<='1';
-- 		elsif MSB_are_equals='1' and LSB_smaller='1' then
-- 			new_is_smaller<='1';
-- 		else
-- 			new_is_smaller<='0';
-- 		end if;
		
		if (MSB_are_equals='1' and LSB_are_equals='1') then
			longs_are_equals<='1';
		else
			longs_are_equals<='0';
		end if;

	end if;
end process;

-------------------------------------------------------------------------------
-- delta sort FSM
-------------------------------------------------------------------------------
jitter<=jitter_acc(34 downto 2);
process(refclk)
begin
	if rising_edge(refclk) then
		if (rst='1') then
			update_local_time<='0';
			last_epoch_used<=(others=>'0');
			last_theta_used<=(others=>'0');
			cnt_since_last_upd<=(others=>'0');
			is_twice_poll_interval<='0';
			state_sort_tuple<=idle;
		else
			case state_sort_tuple is
				when idle =>
					update_local_time<='0';
					current_delta<=successive_tuples(0).delta;
					best_delta_ptr<=0;
					if update_tuple_1ck='1' then
						ptr<=1;
						last_theta_used<=successive_tuples(1).theta; -- save previous offset (dtemp)
						state_sort_tuple<=update_comp_value;
					end if;
				
				when update_comp_value =>
					jitter_acc<=(others=>'0');
					update_cnt<=0;
					candidate_delta<=successive_tuples(ptr).delta;
					state_sort_tuple<=wait_result;

				when wait_result =>
					if update_cnt=comp_dly-1 then
						if new_is_smaller='1' and successive_tuples(ptr).is_valid='1' then
							best_delta_ptr<=ptr;
							current_delta<=successive_tuples(ptr).delta;
						end if;

						if ptr=7 then
							state_sort_tuple<=update_epoch_value;
						else
							ptr<=ptr+1;
							state_sort_tuple<=update_comp_value;
						end if;
					else
						update_cnt<=update_cnt+1;
					end if;
				
				when update_epoch_value =>
					update_cnt<=0;
					candidate_delta<=successive_tuples(best_delta_ptr).epoch;
					current_delta<=last_epoch_used;
					state_sort_tuple<=wait_epoch_result;

				when wait_epoch_result =>
					if update_cnt=comp_dly-1 then
						if (new_is_smaller='0' and longs_are_equals='0') or in_init1='1'then -- we look for a later than last used date, i.e. a larger date, or all the new data
							ptr<=0;
							if in_init1='1' and successive_tuples(0).is_valid='1'then
								best_delta_ptr<=0; --use newest value during this phase whatever the delta
							else
								cnt_since_last_upd<=cnt_since_last_upd+1; -- increment if not kept
								state_sort_tuple<=idle;
							end if;
							state_sort_tuple<=calc_diff_i; -- continue filtering algorithm
						else
							if cnt_since_last_upd=1 then
								is_twice_poll_interval<='1';
							end if;
							cnt_since_last_upd<=cnt_since_last_upd+1; -- increment if not kept
							state_sort_tuple<=idle;
						end if;
					else
						update_cnt<=update_cnt+1;
					end if;

				when calc_diff_i =>
					update_cnt<=0;
					if successive_tuples(ptr).is_valid='1' then -- assumes theta below 0.5s
						diff_i<=successive_tuples(best_delta_ptr).theta(31 downto 0)-successive_tuples(ptr).theta(31 downto 0);
					else
						diff_i<=MAX_DISP;
					end if;
					state_sort_tuple<=put_abs;
				
				when put_abs =>
					if diff_i(31)='1' then --negative => take opposite
						diff_i<=not(diff_i)+1;
					end if;
					state_sort_tuple<=accumulate_jitter;

				when accumulate_jitter=>
					jitter_acc<=jitter_acc+diff_i;
					sub_op_1<=successive_tuples(best_delta_ptr).theta;
					sub_op_2<=last_theta_used;

					if ptr=7 then
						state_sort_tuple<=compute_theta_diff;
					else
						ptr<=ptr+1;
						state_sort_tuple<=calc_diff_i;
					end if;

				when compute_theta_diff=>
					if update_cnt=comp_dly-1 then
						update_cnt<=0;
						if sub_result(63)='1' then --negative => recompute
							sub_op_1<=last_theta_used;
							sub_op_2<=successive_tuples(best_delta_ptr).theta;
						else
							state_sort_tuple<=compare_jitter;
						end if;
					else
						update_cnt<=update_cnt+1;
					end if;

				when compare_jitter=>
					--jitter divided by 8 multiplied by 2 => divided by 4
					-- theta_diff = new_theta-previous_theta
 
					-- apply correction if theta_diff is below jitter or |theta_diff| larger than 1 second ot still in_init1
					if ((sub_result(31 downto 0)<=jitter or sub_result(63 downto 31)/=X"0000_0000") and is_twice_poll_interval='1') or in_init1='1' then
-- 					if true then
						new_theta<=successive_tuples(best_delta_ptr).theta;
						new_delta<=successive_tuples(best_delta_ptr).delta;
						last_epoch_used<=successive_tuples(best_delta_ptr).epoch;
						reference_time<=successive_tuples(best_delta_ptr).epoch;

						n<=cnt_since_last_upd;
						cnt_since_last_upd<=(others=>'0');
						is_twice_poll_interval<='0';
						update_local_time<='1';
					else
						if cnt_since_last_upd=1 then
							is_twice_poll_interval<='1';
						end if;
						cnt_since_last_upd<=cnt_since_last_upd+1; -- increment if not kept
					end if;
					state_sort_tuple<=idle;

			end case;
		end if;
	end if;
end process;
-------------------------------------------------------------------------------
-- subtracter for theta diff
-------------------------------------------------------------------------------
lbl_subtracter : ntp_addsub generic map(
	is_adder=>false,
	size=>64
	)
port map(
	clk=>refclk,

	op_1=>sub_op_1,
	op_2=>sub_op_2,

	result=>sub_result
	);

end RTL;
