-- ntp_udp_rx
--
-- manage the communication with the UDP layer.
-- Use local time and provides T'1,T'2, T1 and T2
--
-- poll interval set to 128s (exponent of 7)
-- timeout value is equal to the poll interval
--
-- Olivier Bourrion, May 2015

library ieee;
use ieee.std_logic_1164.all;
use IEEE.NUMERIC_STD.ALL;

use work.pack_ipv4_types.all;
use work.pack_axi.all;
use work.pack_arp_types.all;

entity ntp_udp_rx is 
port(
	mac_clk: in std_logic; -- Ethernet MAC clock (125MHz)
	rst_macclk: in std_logic; -- MAC clock domain sync reset

	-- UDP RX signals
    udp_rx_start : in std_logic;       -- indicates receipt of udp header
    udp_rxo      : in udp_rx_type;

	local_time : in std_logic_vector(63 downto 0); -- local time
	local_time_at_rcv : out std_logic_vector(63 downto 0); -- local time at datagram rcv time

	rcv_ntp_ack: out std_logic;
	origin_time : out std_logic_vector(63 downto 0); -- rcvd OT field
	receive_time : out std_logic_vector(63 downto 0); -- rcvd RT field
	Transmit_time : out std_logic_vector(63 downto 0) -- rcvd TT field
	);
end ntp_udp_rx;

architecture RTL of ntp_udp_rx is
-- RX side
type def_state_ntp_rx is (idle,rcv_ntp,wait_end);
signal state_ntp_rx : def_state_ntp_rx;
signal rx_count : natural range 0 to 63;
signal save_T2 : std_logic;

begin
-------------------------------------------------------------------------------
-- ntp RX state machine
-------------------------------------------------------------------------------
process(mac_clk)
begin
	if rising_edge(mac_clk) then
		if save_T2='1' then
			local_time_at_rcv<=local_time;
		end if;

		if rst_macclk='1' then
			rcv_ntp_ack<='0';
			state_ntp_rx<=idle;
		else
			case state_ntp_rx is
				when idle =>
					rx_count<=0;
					rcv_ntp_ack<='0';
					if udp_rxo.hdr.is_valid ='1' then
						-- NTP port and not a broadcast packet (no willing to use those)
						if udp_rxo.hdr.dst_port=std_logic_vector(to_unsigned(123,16)) and udp_rxo.hdr.is_broadcast='0' then
							save_T2<='1';
							state_ntp_rx<=rcv_ntp;
						end if;
					else
						save_T2<='0';
					end if;

				when rcv_ntp =>
					save_T2<='0';
					if udp_rxo.data.data_in_valid='1' then
						rx_count<=rx_count+1;
						case rx_count is
							when 24 => origin_time(63 downto 56) <= udp_rxo.data.data_in; -- OT field
							when 25 => origin_time(55 downto 48) <= udp_rxo.data.data_in; -- OT field
							when 26 => origin_time(47 downto 40) <= udp_rxo.data.data_in; -- OT field
							when 27 => origin_time(39 downto 32) <= udp_rxo.data.data_in; -- OT field

							when 28 => origin_time(31 downto 24) <= udp_rxo.data.data_in; -- OT field
							when 29 => origin_time(23 downto 16) <= udp_rxo.data.data_in; -- OT field
							when 30 => origin_time(15 downto 8)  <= udp_rxo.data.data_in; -- OT field
							when 31 => origin_time(7 downto 0)   <= udp_rxo.data.data_in; -- OT field
							
							when 32 => receive_time(63 downto 56) <= udp_rxo.data.data_in; -- RT field
							when 33 => receive_time(55 downto 48) <= udp_rxo.data.data_in; -- RT field
							when 34 => receive_time(47 downto 40) <= udp_rxo.data.data_in; -- RT field
							when 35 => receive_time(39 downto 32) <= udp_rxo.data.data_in; -- RT field

							when 36 => receive_time(31 downto 24) <= udp_rxo.data.data_in; -- RT field
							when 37 => receive_time(23 downto 16) <= udp_rxo.data.data_in; -- RT field
							when 38 => receive_time(15 downto 8)  <= udp_rxo.data.data_in; -- RT field
							when 39 => receive_time(7 downto 0)   <= udp_rxo.data.data_in; -- RT field
							
							when 40 => Transmit_time(63 downto 56) <= udp_rxo.data.data_in; -- TT field
							when 41 => Transmit_time(55 downto 48) <= udp_rxo.data.data_in; -- TT field
							when 42 => Transmit_time(47 downto 40) <= udp_rxo.data.data_in; -- TT field
							when 43 => Transmit_time(39 downto 32) <= udp_rxo.data.data_in; -- TT field

							when 44 => Transmit_time(31 downto 24) <= udp_rxo.data.data_in; -- TT field
							when 45 => Transmit_time(23 downto 16) <= udp_rxo.data.data_in; -- TT field
							when 46 => Transmit_time(15 downto 8)  <= udp_rxo.data.data_in; -- TT field
							when 47 => 
								Transmit_time(7 downto 0)   <= udp_rxo.data.data_in; -- TT field
								if udp_rxo.data.data_in_last='1' then
									rcv_ntp_ack<='1';
									state_ntp_rx<=idle;
								else
								   state_ntp_rx<=wait_end;
								end if;

							when others => null;
						end case;
					end if;
				
				when wait_end =>
					if udp_rxo.data.data_in_valid='1' and  udp_rxo.data.data_in_last='1' then
						rcv_ntp_ack<='1';
						state_ntp_rx<=idle;
					end if;
				
			end case;
		end if;
	end if;
end process;
end RTL;
