----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 		Peter Fall, rewrite by Olivier Bourrion (fusion of process, sync between tx and RX)
-- 
-- Create Date:    5 June 2011 
-- Design Name: 
-- Module Name:    UDP_RX - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--              handle simple UDP RX
--              doesnt check the checsum
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Revision 0.02 - Improved error handling
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.NUMERIC_STD.all;
use work.pack_axi.all;
use work.pack_ipv4_types.all;

entity UDP_RX is
  port (
    -- UDP Layer signals
    udp_rx_start : out std_logic;       -- indicates receipt of udp header
    udp_rxo      : out udp_rx_type;
    -- system signals
    clk          : in  std_logic;
    reset        : in  std_logic;
    -- IP layer RX signals
    ip_rx_start  : in  std_logic;       -- indicates receipt of ip header
    ip_rx        : in  ipv4_rx_type
    );                  
end UDP_RX;

architecture Behavioral of UDP_RX is
type rx_state_type is (IDLE, UDP_HDR, USER_DATA, WAIT_END);

-- state variables
signal rx_state         : rx_state_type;
signal rx_count         : unsigned (15 downto 0);
signal data_len         : std_logic_vector (15 downto 0);  -- user data length captured from input

-- IP datagram header format
--
--      0          4          8                      16      19             24                    31
--      --------------------------------------------------------------------------------------------
--      |              source port number            |              dest port number               |
--      |                                            |                                             |
--      --------------------------------------------------------------------------------------------
--      |                length (bytes)              |                checksum                     |
--      |          (header and data combined)        |                                             |
--      --------------------------------------------------------------------------------------------
--      |                                          Data                                            |
--      |                                                                                          |
--      --------------------------------------------------------------------------------------------
--      |                                          ....                                            |
--      |                                                                                          |
--      --------------------------------------------------------------------------------------------
begin
udp_rxo.hdr.data_length <= data_len;

rx_sequential : process (clk)
begin
    if rising_edge(clk) then
		if reset = '1' then
			-- reset state variables
			rx_state         <= IDLE;
			rx_count         <= x"0000";
			udp_rxo.hdr.src_port    <= (others => '0');
			udp_rxo.hdr.is_valid    <= '0';
			udp_rxo.hdr.rx_error    <= '0';
			udp_rxo.hdr.dst_port    <= (others => '0');
			data_len         <= (others => '0');
			udp_rx_start	<= '0';
			udp_rxo.hdr.src_ip_addr      <= (others => '0');
		else
			case rx_state is
				when IDLE =>
					udp_rx_start <= '0';
					udp_rxo.hdr.is_valid    <= '0';
					udp_rxo.data.data_in       <= (others => '0');
					udp_rxo.data.data_in_valid <= '0';
					udp_rxo.data.data_in_last  <= '0';
					udp_rxo.hdr.rx_error    <= '0';
					if ip_rx.data.data_in_valid = '1' then
						if ip_rx.hdr.protocol = x"11" then -- UDP protocol
							rx_count <= rx_count + 1; 
							udp_rxo.hdr.src_ip_addr <= ip_rx.hdr.src_ip_addr;
							udp_rxo.hdr.src_port(15 downto 8) <= ip_rx.data.data_in;
							udp_rxo.hdr.is_broadcast <= ip_rx.hdr.is_broadcast;
							rx_state <= UDP_HDR;
						else -- non-UDP protocol - ignore this pkt
							rx_count <= x"0000";
							rx_state <= WAIT_END;
						end if;
					end if;

				when UDP_HDR =>
					if ip_rx.data.data_in_valid = '1' then
						if rx_count = x"0007" then
							rx_count<= x"0001";
							rx_state <= USER_DATA;
						else
							rx_count <= rx_count + 1; 
						end if;
												
						if ip_rx.data.data_in_last = '1' then -- handle early frame termination
							rx_state <= IDLE;
						else
							case rx_count is
								when x"0000" => udp_rxo.hdr.src_port(15 downto 8) <= ip_rx.data.data_in; 
								when x"0001" => udp_rxo.hdr.src_port(7 downto 0)  <= ip_rx.data.data_in;
								when x"0002" => udp_rxo.hdr.dst_port(15 downto 8) <= ip_rx.data.data_in; 
								when x"0003" => udp_rxo.hdr.dst_port(7 downto 0)  <= ip_rx.data.data_in;

								when x"0004" => 
									data_len (15 downto 8) <= ip_rx.data.data_in;
									data_len (7 downto 0)  <= x"00";
								when x"0005" => 
									data_len <= std_logic_vector(unsigned(data_len(15 downto 8) & ip_rx.data.data_in) - 8);
									udp_rxo.hdr.is_valid    <= '1';-- header values are now valid, although the pkt may not be for us

								when x"0006" =>                       -- ignore checksum values
								when x"0007" => udp_rx_start <= '1';  -- indicate frame received
								when others =>  -- ignore other bytes in udp header                                                                              
							end case;
						end if;
					end if;
				
				when USER_DATA =>
					udp_rxo.data.data_in       <= ip_rx.data.data_in;
					udp_rxo.data.data_in_valid <= ip_rx.data.data_in_valid;
					if ip_rx.hdr.last_error_code/= RX_EC_NONE then -- packet in error
						udp_rxo.hdr.rx_error    <= '1';
					end if;

					if ip_rx.data.data_in_valid = '1' then
						if rx_count = unsigned(data_len) then
							udp_rx_start <= '0';
							rx_count    <= x"0000";
							udp_rxo.data.data_in_last<= '1';
							if ip_rx.data.data_in_last = '1' then
								rx_state    <= IDLE;
							else
								rx_state    <= WAIT_END; -- probably an error
							end if;
						else
							rx_count<= rx_count+1;
							if ip_rx.data.data_in_last = '1' then -- probably an error
								rx_state <= IDLE;
								udp_rxo.data.data_in_last<= '1';
							end if;
						end if;
					end if;

				when WAIT_END =>
					if ip_rx.hdr.last_error_code/= RX_EC_NONE then -- packet in error
						udp_rxo.hdr.rx_error    <= '1';
					end if;
					if ip_rx.data.data_in_valid = '1' then
						if ip_rx.data.data_in_last = '1' then
						  rx_state <= IDLE;
						end if;
					end if;
			end case;
		end if;
	end if;
end process;
end Behavioral;

