----------------------------------------------------------------------------------
-- Company: 
-- Engineer:            Olivier Bourrion
-- 
-- Create Date:    12:00:04 2 oct 2014
-- Design Name: 
-- Module Name:    arp_store
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description:
--              ARP storage table using block ram with lookup based on IP address
--              implements upto XX entries with sequential search
--              uses round robin overwrite when full 
--
--              store may take a number of cycles and the request is latched
--              lookup may take a number of cycles. Assumes that request signals remain valid during lookup
--				round robien replacement, assumes ARP table will never be full (few connections)
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.NUMERIC_STD.all;
use ieee.std_logic_unsigned.all;

entity arp_store is
generic (
    MAX_ARP_ENTRIES : integer := 16 -- max entries in the store
    );
port (
    -- system signals
    clk         : in  std_logic;
    reset       : in  std_logic;

    -- arp lookup signals
    req_ip    : in  std_logic_vector(31 downto 0);
	do_lookup : in std_logic;
	lookup_done : out std_logic;
    rslt_mac    : out  std_logic_vector(47 downto 0);
	req_found : out std_logic;

	-- arp table write signals
    wr_ip    : in  std_logic_vector(31 downto 0);
    wr_mac    : in  std_logic_vector(47 downto 0);
	insert_arp : in std_logic;
	insert_done : out std_logic;

    -- control and status signals
    clear_store : in  std_logic;              -- erase all entries
    entry_count : out std_logic_vector(7 downto 0)   -- how many entries currently in store
    );
end arp_store;

architecture RTL of arp_store is
type ip_ram_t is array (0 to MAX_ARP_ENTRIES-1) of std_logic_vector(31 downto 0);
type mac_ram_t is array (0 to MAX_ARP_ENTRIES-1) of std_logic_vector(47 downto 0);
type arp_mode_t is (RD, WR);
type st_state_t is (IDLE, PAUSE, SEARCH, FOUND, NOT_FOUND);
signal st_state        : st_state_t;

signal ip_ram : ip_ram_t;
signal mac_ram : mac_ram_t;
signal ip_ram_out : std_logic_vector(31 downto 0);
signal mac_ram_out : std_logic_vector(47 downto 0);
signal ip2lookup : std_logic_vector(31 downto 0);

signal arp_mode : arp_mode_t;
-- signal good_ip_pointed : std_logic;
signal wr_cache : std_logic;
signal incr_cache : std_logic;

signal wr_add		   : natural range 0 to MAX_ARP_ENTRIES-1;
signal next_wr_add     : natural range 0 to MAX_ARP_ENTRIES-1;
signal num_entries     : natural range 0 to MAX_ARP_ENTRIES;
signal test_entry      : natural range 0 to MAX_ARP_ENTRIES;
signal rd_add		   : natural range 0 to MAX_ARP_ENTRIES-1;
signal used_rd_add     : natural range 0 to MAX_ARP_ENTRIES-1;

begin
-------------------------------------------------------------------------------
-- Memory block (inferring is possible)
-------------------------------------------------------------------------------
process (clk)
begin
	if rising_edge(clk) then
		-- write side
		if wr_cache='1' then
			ip_ram(wr_add)<=wr_ip;
			mac_ram(wr_add)<=wr_mac;
		end if;
		
		-- read side
		ip_ram_out<=ip_ram(rd_add);
		mac_ram_out<=mac_ram(rd_add);
	end if;
end process;
-------------------------------------------------------------------------------
-- CACHE management FSM
-------------------------------------------------------------------------------
insert_done<=wr_cache;
process (clk)
begin
	if rising_edge(clk) then
		used_rd_add<=rd_add;

-- 		if req_ip=ip_ram_out then
-- 			good_ip_pointed<='1';
-- 		else
-- 			good_ip_pointed<='0';
-- 		end if;

		if reset='1' then
			rd_add<=0;
			wr_cache   <= '0';
			incr_cache<= '0';
			lookup_done<='0';
			st_state <= IDLE;
		else
			case st_state is
				when IDLE =>
					wr_cache   <= '0';
					incr_cache<= '0';
					req_found<='0';
					rslt_mac<=mac_ram_out;
					test_entry<=0;
					if insert_arp = '1' then -- for write search start from 0
						rd_add<=0;
						arp_mode<=WR;
						ip2lookup<=wr_ip;
						st_state <= PAUSE;
					elsif do_lookup = '1' then
						arp_mode<=RD;
						ip2lookup<=req_ip;
-- 						if good_ip_pointed='1' then --for boosting search if already pointed
						if req_ip=ip_ram_out then --for boosting search if already pointed
							st_state<=FOUND;
							lookup_done<='1';
							req_found<='1';
						else -- not existing scan table
							if num_entries=0 then
								st_state<=NOT_FOUND;
								lookup_done<='1';
								req_found<='0';
							else
								rd_add<=0;
								st_state <= PAUSE;
							end if;
						end if;
					end if;

				when PAUSE =>
					if test_entry/=num_entries then
						test_entry<=test_entry+1;
						rd_add<=rd_add+1;
					end if;
					st_state<= SEARCH;
					
				when SEARCH =>
					rslt_mac<=mac_ram_out;
					if ip2lookup = ip_ram_out then -- found
						if arp_mode=RD then
							lookup_done<='1';
							req_found<='1';
						end if;
						rd_add<=used_rd_add;
						wr_add <= used_rd_add;
						st_state <= FOUND;
					else
						if test_entry = num_entries then-- reached end of entry table
							if arp_mode=RD then
								lookup_done<='1';
							end if;
							st_state <= NOT_FOUND;
						else
							test_entry<=test_entry+1;
						end if;

						if rd_add /=MAX_ARP_ENTRIES-1 then
							rd_add<=rd_add+1;
						end if ;
					end if;

				when FOUND =>
					req_found<='0';
					lookup_done<='0';
					if arp_mode = WR then -- update existing ARP
						wr_cache   <= '1';
						st_state <= IDLE;
					elsif do_lookup = '0' then   -- wait in this state until request de-asserted
						st_state <= IDLE;
					end if;

				when NOT_FOUND =>
					lookup_done<='0';
					if arp_mode = WR then
						wr_cache<= '1';
						incr_cache<= '1';
						wr_add<=next_wr_add;
						st_state <= IDLE;
					elsif do_lookup = '0' then   -- wait in this state until request de-asserted
						st_state <= IDLE;
					end if;
			end case;
		end if;
	end if;
end process;

-------------------------------------------------------------------------------
-- CACHE
-------------------------------------------------------------------------------
entry_count<= std_logic_vector(to_unsigned(num_entries, 8));
process (clk)
begin
	if rising_edge(clk) then
		if clear_store='1' or reset='1' then
			num_entries<=0;
			next_wr_add<=0;
		else
			if incr_cache='1' then
				if num_entries/=MAX_ARP_ENTRIES then
					num_entries<=num_entries+1;
				end if;

				if next_wr_add=MAX_ARP_ENTRIES-1 then -- round robin replacement (best would be Least Recently used)
					next_wr_add<=0;
				else
					next_wr_add<=next_wr_add+1;
				end if;
			end if;
		end if;
	end if;
end process;

end RTL;
